import networkx as nx
from mmodel.draw import draw_graph
from copy import deepcopy
from mmodel.filter import subnodes_by_inputs, subnodes_by_outputs
from mmodel.utility import (
    modelgraph_signature,
    modelgraph_returns,
    replace_subgraph,
    modify_node,
)
from mmodel.metadata import nodeformatter, format_metadata, textwrap80
from mmodel.parser import node_parser


class ModelGraph(nx.DiGraph):

    """Create model graphs.

    ModelGraph inherits from `networkx.DiGraph()`, which has all `DiGraph`
    methods.

    The class adds the "type" attribute to the graph attribute. The factory method
    returns a copy of the dictionary. It is equivalent to
    ``{"type": "ModelGraph"}.copy()`` when called.

    The additional graph operations are added:
    - add_grouped_edges and set_node_objects.
    - Method ``add_grouped_edges``, cannot have both edges list.
    - Method ``set_node_object`` updates nodes with the node callable "func" and output.
    - The method adds callable signature 'sig' to the node attribute.
    """

    # Add the default parser to the graph attribute dictionary.
    # To change the parser, subclass ModelGraph and change the attribute.
    graph_attr_dict_factory = {"type": "ModelGraph", "parser": node_parser}.copy

    def set_node_object(
        self,
        node: str,
        func: callable,
        output: str = None,
        inputs: list = None,
        modifiers: list = None,
        **kwargs,
    ):
        """Add or update the functions of an existing node.

        In the end, the edge attributes are re-determined
        Modifiers are applied directly onto the node. The parser checks the
        function type and returns (at least) three dictionary entries:
        _func, functype, doc.
        """

        node_dict = self.nodes[node]
        parser = self.graph["parser"]

        modifiers = modifiers or list()
        attr_dict = parser(node, func, output, inputs, modifiers, **kwargs)
        node_dict.update(attr_dict)
        node_dict.update(kwargs)

        self.update_graph()

    def set_node_objects_from(self, node_objects: list):
        """Update the functions of existing nodes.

        The method is the same as adding a node object.
        """

        for node_obj in node_objects:
            # unzipping works for input with or without modifiers
            self.set_node_object(*node_obj)

    def add_edge(self, u_of_edge, v_of_edge, **attr):
        """Modify add_edge to update the edge attribute in the end."""

        super().add_edge(u_of_edge, v_of_edge, **attr)
        self.update_graph()

    def add_edges_from(self, ebunch_to_add, **attr):
        """Modify add_edges_from to update the edge attributes."""

        super().add_edges_from(ebunch_to_add, **attr)
        self.update_graph()

    def add_grouped_edge(self, u, v):
        """Add linked edge.

        For mmodel, a group edge (u, v) allows u or v
        to be a list of nodes. Represents several nodes
        flowing into one node or the other way around.
        """

        if isinstance(u, list) and isinstance(v, list):
            raise Exception("only one edge node can be a list")

        # use add edges from to run less update graph
        # currently a compromise
        if isinstance(u, list):
            self.add_edges_from([(_u, v) for _u in u])
        elif isinstance(v, list):
            self.add_edges_from([(u, _v) for _v in v])
        else:  # neither is a list
            self.add_edge(u, v)

    def add_grouped_edges_from(self, group_edges: list):
        """Add edges from grouped values."""

        for u, v in group_edges:
            self.add_grouped_edge(u, v)

    def update_graph(self):
        """Update edge attributes based on node objects and edges."""

        for u, v in self.edges:
            # the edge "var" is not defined if the parent node does not
            # have "output" attribute or the child node does not have
            # the parameter

            # extract the parameter dictionary
            v_sig = getattr(self.nodes[v].get("sig", None), "parameters", {})
            if "output" in self.nodes[u] and self.nodes[u]["output"] in v_sig:
                self.edges[u, v]["var"] = self.nodes[u]["output"]

    def _node_metadata_dict(self, node: str, verbose: bool):
        """Return node metadata as a dictionary."""

        node_dict = self.nodes[node]

        short_dict = {
            "node": node,
            "func": node_dict["func"],
            "return": node_dict["output"],
        }

        additional_dict = {
            "functype": node_dict["functype"],
            "modifiers": node_dict["modifiers"],
            "doc": node_dict["doc"],
        }
        if verbose:
            return {**short_dict, **additional_dict}
        else:
            return short_dict

    def node_metadata(
        self, node: str, verbose=True, formatter=nodeformatter, textwrapper=textwrap80
    ):
        """Printout node metadata.

        The metadata includes the node information and the node function
        information.
        """
        str_list = format_metadata(
            self._node_metadata_dict(node, verbose), formatter, textwrapper
        )
        return "\n".join(str_list).rstrip()

    # graph properties
    @property
    def signature(self):
        """Graph signature.

        :rtype: inspect.Signature object
        """
        return modelgraph_signature(self)

    @property
    def returns(self):
        """Graph returns.

        :rtype: list
        """
        return modelgraph_returns(self)

    # graph operations
    def subgraph(self, nodes=None, inputs=None, outputs=None):
        """Extract subgraph by nodes, inputs, and output.

        If multiple parameters are specified, the result is a union
        of the selection. The subgraph is a deep copy of the original graph.
        The behavior is different from the parent class method, where the subgraph
        returns a view of the original graph.
        """

        nodes = nodes or []
        node_inputs = subnodes_by_inputs(self, inputs or [])
        node_outputs = subnodes_by_outputs(self, outputs or [])

        # convert nodes to list because the parent class method accepts generator
        # for nodes.
        # may consider not using the same name as the parent class to avoid collision
        subgraph_nodes = set(list(nodes) + node_inputs + node_outputs)  # unique nodes

        return super().subgraph(subgraph_nodes).deepcopy()

    def replace_subgraph(
        self, subgraph, name, func, output=None, inputs=None, modifiers=None
    ):
        """Replace subgraph with a node."""
        return replace_subgraph(self, subgraph, name, func, output, inputs, modifiers)

    def modify_node(
        self, node, func=None, output=None, inputs=None, modifiers=None, inplace=False
    ):
        """Modify node attributes."""
        return modify_node(self, node, func, output, inputs, modifiers, inplace)

    def draw(self, style="verbose", export=None):
        """Draw the graph.

        Draws the default styled graph.

        :param str style: there are three styles, plain, short, and verbose.
            Plain shows nodes only, short shows part of the metadata, and
            long shows all the metadata.
        :param str export: filename to save the graph as. The file extension
            is needed.
        """

        return draw_graph(self, str(self), style, export)

    def deepcopy(self):
        """Deepcopy graph.

        The graph.copy method is a shallow copy. Deepcopy creates copy for the
        attributes dictionary.
        `graph.copy<https://networkx.org/documentation/stable/reference/classes
        /generated/networkx.Graph.copy.html>_`

        However, for subgraphs, deepcopy is incredibly inefficient because
        subgraph contains '_graph', which stores the original graph.
        An alternative method is to copy the code from the copy method,
        but use deepcopy for the items.

        The parser is redefined in the new graph.
        """

        G = self.__class__()
        G.graph.update(deepcopy(self.graph))
        G.add_nodes_from((n, deepcopy(d)) for n, d in self._node.items())
        G.add_edges_from(
            (u, v, deepcopy(datadict))
            for u, nbrs in self._adj.items()
            for v, datadict in nbrs.items()
        )

        return G



import graphviz
from copy import deepcopy
from mmodel.metadata import textwrap50, nodeformatter
import re

DEFAULT_SETTINGS = {
    "graph_attr": {
        "labelloc": "t",
        "labeljust": "l",
        "splines": "ortho",
        "ordering": "out",
    },
    "node_attr": {"shape": "box"},
}


def replace_label(label):
    """Format label for graphviz.

    The function replaces newlines with the graphviz left aligned line break.
    However if the ``\n`` is escaped, change it to ``\\\\n``.
    """
    return re.sub(r"(?<!\\)\n", r"\\l", label).replace("\\n", "\\\\n") + r"\l"


def draw_graph(
    G, label, style, export=None, formatter=nodeformatter, textwrapper=textwrap50
):
    """Draw a detailed graph with options.

    :param str name: name of the graph
    :param str label: title of the graph
    :param str style: there are three valid styles, plain
        short and verbose. Each style corresponds to node-only,
        function-only, and detailed note metadata graph
    :param str export: filename to export to
    """

    label = label.replace("\n", r"\l") + r"\l"
    settings = deepcopy(DEFAULT_SETTINGS)
    settings["graph_attr"].update({"label": label})

    dot_graph = graphviz.Digraph(name=G.name, **settings)

    if style == "plain":
        for node in G.nodes:
            dot_graph.node(node)

        for u, v in G.edges:
            dot_graph.edge(u, v)

    else:
        if style == "short":
            for node, ndict in G.nodes(data=True):
                if "func" in ndict:
                    metadata = G.node_metadata(node, False, formatter, textwrapper)
                    nlabel = replace_label(metadata)
                else:
                    nlabel = node

                dot_graph.node(node, label=nlabel)

        elif style == "verbose":
            for node, ndict in G.nodes(data=True):
                if "func" in ndict:
                    metadata = G.node_metadata(node, True, formatter, textwrapper)
                    nlabel = replace_label(metadata)
                else:
                    nlabel = node

                dot_graph.node(node, label=nlabel)

        else:
            raise Exception(
                f"Invalid style {repr(style)}: must be one of plain, short, or verbose."
            )

        for u, v, edict in G.edges(data=True):

            if "var" in edict:
                xlabel = edict["var"]
            else:
                xlabel = ""

            dot_graph.edge(u, v, xlabel=xlabel)

    if export:
        dot_graph.render(outfile=export)

    return dot_graph



__all__ = ["modifier_shortcut"]

def modifier_shortcut(model, modifier_dict, name=None):
    """Shortcut for quickly adding modifiers to the model."""

    name = name or model.name
    G = model.graph  # copy of the graph
    for node, modifier_list in modifier_dict.items():
        modifiers = G.nodes[node]["modifiers"]
        modifiers = modifiers + modifier_list
        G.modify_node(node, modifiers=modifiers, inplace=True)

    return model.edit(name=name, graph=G)



__all__ = [
    "loop_input",
    "zip_loop_inputs",
    "redefine_signature",
    "redefine_pos_signature",
    "bind_signature",
    "profile_time",
]


from functools import wraps
import inspect
from mmodel.utility import parse_input, parse_parameters


def loop_input(parameter: str):
    """Modify function to iterate one given parameter.

    :param list parameter: target parameter to loop
    """

    def loop(func):
        @wraps(func)
        def loop_wrapped(**kwargs):

            loop_values = kwargs.pop(parameter)

            try:
                # make sure the parameter input is iterable
                iter(loop_values)
            except TypeError:
                raise Exception(f"{parameter} value is not iterable")

            return [func(**kwargs, **{parameter: value}) for value in loop_values]

        return loop_wrapped

    loop.metadata = f"loop_input({repr(parameter)})"
    return loop


def zip_loop_inputs(parameters: list):
    """Modify function to iterate the parameters pairwise.

    :param list parameters: list of the parameter to loop
        only one parameter is allowed. If the string of parameters is
        provided, the parameters should be delimited by ", ".
    """

    def zip_loop(func):
        @wraps(func)
        def loop_wrapped(**kwargs):

            loop_values = [kwargs.pop(param) for param in parameters]

            result = []
            for value in zip(*loop_values):  # unzip the values

                loop_value_dict = dict(zip(parameters, value))
                rv = func(**kwargs, **loop_value_dict)
                result.append(rv)

            return result

        return loop_wrapped

    zip_loop.metadata = f"zip_loop({repr(parameters)})"
    return zip_loop


def redefine_signature(parameters: list):
    """Replace node object signature.

    :param list parameters: signature parameters to replace the original
        signature. The parameters are assumed to be "POSITIONAL_OR_KEYWORD", and no
        default values are allowed. The signature will replace the original signature
        in order.

    .. Note::

        The modifier does not work with functions that have positional-only
        input parameters; or functions do not have a signature
        (built-in and numpy.ufunc). Use pos_signature_modifier instead.

    """

    def signature_modifier(func):

        sig, param_order, defaultargs = parse_parameters(parameters)

        # if there's "kwargs" in the parameter, ignore the parameter
        old_parameters = []
        for name, param in inspect.signature(func).parameters.items():
            if param.kind < 4:
                old_parameters.append(name)

        # once unzipped to return iterators, the original variable returns none
        param_pair = list(zip(old_parameters, param_order))

        @wraps(func)
        def wrapped(**kwargs):
            # assume there are no repeated signature names that are not overlapping
            # replace a, b with b, c is not allowed in the following step
            kwargs.update(defaultargs)
            for old, new in param_pair:
                kwargs[old] = kwargs.pop(new)
            return func(**kwargs)

        wrapped.__signature__ = sig
        return wrapped

    return signature_modifier


def redefine_pos_signature(parameters: list):
    """Replace node object positional signature with keyword arguments.

    For functions that do not have a signature or only allow positional only
    inputs. The modifier is specifically used to change the "signature" of the builtin
    function or numpy.ufunc.

    .. note::

        The modifier is only tested against built-in function or numpy.ufunc.
    """

    def pos_signature_modifier(func):

        sig, param_order, defaultargs = parse_parameters(parameters)

        @wraps(func)
        def wrapped(**kwargs):
            kwargs.update(defaultargs)
            # extra the variables in order
            inputs = [kwargs[key] for key in param_order]

            return func(*inputs)

        wrapped.__signature__ = sig
        return wrapped

    return pos_signature_modifier


def bind_signature(func):
    """Add parameter binding and checking for function.

    The additional wrapper is unnecessary, but to keep a consistent
    modifier syntax. The modifier can be used on wrapped functions
    that do not have a parameter binding step (ones that only allow
    keyword arguments).

    The parse_input method binds the input args and kwargs and fills
    default values automatically. The resulting function behaves the
    same as a python function.
    """

    sig = inspect.signature(func)

    @wraps(func)
    def wrapped(*args, **kwargs):

        parsed_kwargs = parse_input(sig, *args, **kwargs)

        return func(**parsed_kwargs)

    return wrapped


def format_time(dt, precision):
    """Format time in seconds to a human-readable string."""

    units = {"s": 1.0, "ms": 1e-3, "us": 1e-6, "ns": 1e-9}
    for unit, scale in units.items():
        if dt >= scale:
            return f"{dt / scale:.{precision}f} {unit}"


def profile_time(number=1, repeat=1, verbose=False, precision=2):
    """Profile the execution time of a function.

    The modifier behaves similarly to the timeit module. However,
    the modifier does not suppress garbage collection during the function
    execution, therefore, the result might be slightly different.
    """
    import timeit

    timer = timeit.default_timer

    def timeit_modifier(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            time_list = []

            for _ in range(repeat):
                t0 = timer()
                for _ in range(number):
                    result = func(*args, **kwargs)
                t1 = timer()
                time_list.append((t1 - t0) / number)

            if verbose:
                print(f"{func.__name__} - raw times: {time_list}")
            else:
                term = "loops" if number > 1 else "loop"
                min_time = format_time(min(time_list), precision)
                print(
                    f"{func.__name__} - {number} {term}, "
                    f"best of {repeat}: {min_time} per loop"
                )
            return result

        return wrapped

    timeit_modifier.metadata = (
        f"profile_time(number={number}, repeat={repeat}, verbose={verbose})"
    )
    return timeit_modifier



"""Handle node, graph and model metadata.

The two core types of functions are formatter and wrapper. The formatter
takes care of the formatting of the metadata, and the wrapper takes care
of the wrapping of the metadata.
"""

import inspect
from textwrap import TextWrapper


class MetaDataFormatter:
    """Metadata Formatter."""

    def __init__(self, formatter: dict, metaorder: list = None):
        """Initiate the formatter."""

        self.formatter = formatter
        self.metaorder = metaorder

    def __call__(self, metadata):
        """Convert metadata dictionary to string.

        The process returns a list of metadata by lines.
        If the formatter is not found in the formatter dictionary,
        the default string output is "key: value".

        :param dict metadata: metadata dictionary
        :param list metaorder: metadata key order, entry is None if linebreak needed.
            Defaults to dictionary key order.
        """
        metaorder = self.metaorder if self.metaorder is not None else metadata.keys()
        metadata_str_list = []
        for key in metaorder:
            if key is None:
                metadata_str_list.append("")
                continue
            elif key not in metadata:
                continue
            if key in self.formatter:
                metadata_str_list.extend(self.formatter[key](key, metadata[key]))
            else:
                metadata_str_list.extend([f"{key}: {metadata[key]}"])

        return metadata_str_list


def format_func(key, value):
    """Format the metadata value that has a function.

    The key name is not shown in the string output.
    The result is func(args1, args2, ...)."""

    return [f"{value.__name__}{inspect.signature(value)}"]


def format_list(key, value):
    """Format the metadata value that is a list."""

    if value:
        str_list = [f"{key}:"]
        for v in value:
            str_list.append(f"\t- {v}")
        return str_list
    else:
        return []


def modifier_metadata(closure):
    """Extract metadata from closure, including the name and the arguments.

    The order of extraction:
    1. If the object has the "metadata" attribute defined.
    2. If the closure takes no arguments, the name is the function name.
    3. If the closure takes arguments, the "metadata" attribute is not defined.

    Note::

        inspect.getclosurevars(closure).nonlocals can only parse values
        if the value is used in the closure.
    """

    if hasattr(closure, "metadata"):
        return closure.metadata
    elif not inspect.getclosurevars(closure).nonlocals:
        return closure.__name__

    else:  # closure takes arguments

        # In some rare cases, the closure is a nested function.
        # For example, in the tests, the nested closure reflects the
        # path of the parent function. Here we remove the nested
        # parent function name.

        name = closure.__qualname__.rsplit(".<locals>.")[-2]
        kwargs = inspect.getclosurevars(closure).nonlocals
        kwargs_str = ", ".join(f"{k}={repr(v)}" for k, v in kwargs.items())
        return f"{name}({kwargs_str})"


def format_modifierlist(key, value):
    """Format the metadata that is a list of modifiers.

    The metadata of the modifier is extracted by the modifier_metadata function.
    The resulting list is formatted by the format_list function.
    """

    modifier_str_list = [modifier_metadata(modifier) for modifier in value]

    return format_list(key, modifier_str_list)


def format_dictargs(key, value):
    """Format the metadata value that is a dictionary."""

    if value:
        str_list = [f"{key}:"]

        for k, v in value.items():
            mod_str = f"\t- {k}: {v}"
            str_list.append(mod_str)

        return str_list

    else:
        return []


def format_args(key, value):
    """Format the metadata value that has the value of (func, kwargs).

    The formatter is for the handler metadata.
    """
    if value:
        func, kwargs = value
        return [
            f"{key}: {func.__name__}({', '.join(repr(v) for v in kwargs.values())})"
        ]
    else:
        return []


def format_returns(key, value):
    """Format the metadata value that has a list of returns.

    The formatter is for the returns metadata. If the "returns" value is empty,
    the output is None. If the returns only have 1 value, return the value, else
    return the values separated by commas in a tuple representation.
    """

    return_len = len(value)

    if return_len == 0:
        returns_str = "None"
    elif return_len == 1:
        returns_str = value[0]
    else:
        returns_str = f"({', '.join(value)})"

    return [f"{key}: {returns_str}"]


def format_value(key, value):
    """Format the metadata without displaying the key."""
    if value:
        return [value]
    else:
        return []


def format_obj(key, value):
    """Format the metadata value that is an object.

    Only show the name of the object. This is used for graph objects.
    The object needs to have __name__ or name attribute defined.
    """

    name = getattr(value, "__name__", getattr(value, "name", None))
    if name:
        return [f"{key}: {name}"]
    else:
        return []


modelformatter = MetaDataFormatter(
    {
        "model": format_func,
        "returns": format_returns,
        "graph": format_obj,
        "handler": format_obj,
        "handler args": format_dictargs,
        "modifiers": format_modifierlist,
        "description": format_value,
    },
    [
        "model",
        "returns",
        "graph",
        "handler",
        "handler args",
        "modifiers",
        None,
        "description",
    ],
)

nodeformatter = MetaDataFormatter(
    {
        "node": format_value,
        "func": format_func,
        "modifiers": format_modifierlist,
        "doc": format_value,
    },
    ["node", None, "func", "return", "functype", "modifiers", None, "doc"],
)


def textwrapper(width: int = 80, indent: int = 2):
    """Wrap metadata content.

    The width defaults to 80 characters. The resulting wrapping has no
    initial indentation. The indent parameter is the subsequent indent
    parameter in the wrap function. The tabsize is the same as
    the indent.
    """

    return TextWrapper(
        width=width,
        subsequent_indent=" " * indent,
        replace_whitespace=False,
        expand_tabs=True,
        tabsize=indent,
    )


# Standard textwrapper with 80 characters.
textwrap80 = textwrapper(80, 2)
# shorted textwarpper with 5o characters for nodes.
textwrap50 = textwrapper(50, 2)


def format_metadata(metadata, formatter, textwrapper):
    """Format and wrap the metadata."""

    metadata_list = formatter(metadata)

    if textwrapper:
        metadata_wrapped = []
        for line in metadata_list:
            if line:
                metadata_wrapped.extend(textwrapper.wrap(line))
            else:
                metadata_wrapped.append("")
        return metadata_wrapped

    return metadata_list



from mmodel.utility import parse_input, is_node_attr_defined, is_edge_attr_defined
from mmodel.metadata import modelformatter, textwrap80, format_metadata
from mmodel.draw import draw_graph
import networkx as nx


class Model:
    """Create model callable.

    :param str name: Model name
    :param object graph: ModelGraph instance (digraph)
    :param class handler: Handler class that handles model execution and
        the keyword arguments. The parameter format is (HandlerClass, {})
        By default, the handler takes the graph as the first parameter.
        For additional arguments, add an argument to the dictionary afterward.
    :param list modifiers: modifiers used for the whole graph model executable.
        The parameter is Optional and defaults to an empty list. For each modifier,
        the format is (modifier, {}). All modifiers should have the function as
        the first argument.
    :param str description: model description
    :param list returns: The order of returns of the model; defaults to the
        topological search.
    """

    _model_keys = ["name", "graph", "handler", "modifiers", "description", "returns"]

    def __init__(
        self,
        name,
        graph,
        handler,
        modifiers: list = None,
        description: str = "",
        returns: list = None,
        **kwargs,
    ):

        assert self._is_valid_graph(graph)
        self.name = self.__name__ = name

        # create a copy of the graph
        self._graph = nx.freeze(graph.deepcopy())
        self.returns = returns or self._graph.returns  # tuples
        self.modifiers = modifiers or list()
        self.handler = handler
        self.handler_args = kwargs
        self.description = description

        executor = handler(name, self._graph, self.returns, **kwargs)
        self.execution_order = [node for node, _ in executor.order]

        for mdf in self.modifiers:
            executor = mdf(executor)

        self.__signature__ = executor.__signature__

        # final callable
        self._executor = executor

    def __call__(self, *args, **kwargs):
        """Execute the model.

        The inputs from the keyword arguments are parsed and passed to the
        the handler class.
        """

        # process inputs
        inputs = parse_input(self.__signature__, *args, **kwargs)

        return self._executor(**inputs)

    def _metadata_dict(self, verbose):
        """Return a dictionary with metadata keys."""

        short_dict = {"model": self, "returns": self.returns}

        additonal_dict = {
            "graph": self._graph,
            "handler": self.handler,
            "handler args": self.handler_args,
            "modifiers": self.modifiers,
            "description": self.description,
        }

        if verbose:
            return {**short_dict, **additonal_dict}
        else:
            return short_dict

    def metadata_str(
        self, verbose=True, formatter=modelformatter, textwrapper=textwrap80
    ):
        """Parse metadata string of the Model instance."""
        return format_metadata(self._metadata_dict(verbose), formatter, textwrapper)

    def __str__(self):
        """Output callable information."""

        str_list = self.metadata_str()
        return "\n".join(str_list).rstrip()

    @staticmethod
    def _is_valid_graph(G):
        """Check if the model graph is valid to build an executable.

        ``mmodel`` does not allow cycle graphs, graphs with isolated nodes,
        and all nodes have callable attributes defined.
        The method is bound to the Model class because the features
        are specific to ``Model`` class.
        """

        assert nx.is_directed(G), f"invalid graph ({G.name}): undirected graph."
        assert not nx.recursive_simple_cycles(
            G
        ), f"invalid graph ({G.name}): graph contains cycles."

        assert is_node_attr_defined(G, "func", "callable")
        # the following might occur when the node object is incorrectly constructed
        assert is_node_attr_defined(G, "output")
        assert is_node_attr_defined(G, "sig", "signature")
        assert is_edge_attr_defined(G, "var", "variable")

        return True

    @property
    def graph(self):
        """The graph attribute output a copy of the graph."""
        return self._graph.deepcopy()

    def get_node(self, node):
        """Quick access to node within the model."""

        return self._graph.nodes[node]

    def get_node_func(self, node):
        """Quick access to node base callable within the model.

        The function helps extract the original function within
        the node.
        """

        return self._graph.nodes[node]["_func"]

    def node_metadata(self, *args, **kwargs):
        """View a specific node."""

        return self._graph.node_metadata(*args, **kwargs)

    def draw(self, style="verbose", export=None):
        """Draw the graph of the model.

        Draws the default styled graph.

        :param str style: there are three styles, plain, short, and verbose.
            Plain shows nodes only, short shows part of the metadata, and
            long shows all the metadata.
        :param str export: filename to save the graph as. The file extension
            is needed.
        """

        return draw_graph(self._graph, str(self), style, export)

    def edit(self, **kwargs):
        """Edit components of the model to create a new model.

        It is not recommended to edit the graph component of the model.
        Although it does create a new model, "edit" is for creating
        a new model with the same graph.
        """

        model_dict = {key: getattr(self, key) for key in self._model_keys}
        return self.__class__(**{**model_dict, **kwargs})
